"""
Time zone information
"""

tz_str = '''-1200 Y
-1100 X NUT SST
-1000 W CKT HAST HST TAHT TKT
-0900 V AKST GAMT GIT HADT HNY
-0800 U AKDT CIST HAY HNP PST PT
-0700 T HAP HNR MST PDT
-0600 S CST EAST GALT HAR HNC MDT
-0500 R CDT COT EASST ECT EST ET HAC HNE PET
-0400 Q AST BOT CLT COST EDT FKT GYT HAE HNA PYT
-0300 P ADT ART BRT CLST FKST GFT HAA PMST PYST SRT UYT WGT
-0200 O BRST FNT PMDT UYST WGST
-0100 N AZOT CVT EGT
+0000 Z EGST GMT UTC WET WT
+0100 A CET DFT WAT WEDT WEST
+0200 B CAT CEDT CEST EET SAST WAST
+0300 C EAT EEDT EEST IDT MSK
+0400 D AMT AZT GET GST KUYT MSD MUT RET SAMT SCT
+0500 E AMST AQTT AZST HMT MAWT MVT PKT TFT TJT TMT UZT YEKT
+0600 F ALMT BIOT BTT IOT KGT NOVT OMST YEKST
+0700 G CXT DAVT HOVT ICT KRAT NOVST OMSST THA WIB
+0800 H ACT AWST BDT BNT CAST HKT IRKT KRAST MYT PHT SGT ULAT WITA WST
+0900 I AWDT IRKST JST KST PWT TLT WDT WIT YAKT
+1000 K AEST ChST PGT VLAT YAKST YAPT
+1100 L AEDT LHDT MAGT NCT PONT SBT VLAST VUT
+1200 M ANAST ANAT FJT GILT MAGST MHT NZST PETST PETT TVT WFT
+1330 FJST NZDT
+1130 NFT
+1030 ACDT LHST
+0930 ACST
+0630 CCT MMT
+0545 NPT
+0530 SLT
+0430 AFT IRDT
+0330 IRST
-0230 HAT NDT
-0330 HNT NST NT
-0430 HLV VET
-0930 MART MIT'''

tz = {}
for tz_descr in map(str.split, tz_str.split('\n')):
    tz_offset = tz_descr[0]
    for tz_code in tz_descr[1:]:
        tz[tz_code] = tz_offset
